import logging
from typing import Any, List, Tuple

from pydantic import ValidationError

import datadoc.state as state
from datadoc.frontend.callbacks.utils import (
    find_existing_language_string,
    get_options_for_language,
    update_global_language_state,
)
from datadoc.frontend.fields.DisplayDataset import (
    DISPLAYED_DATASET_METADATA,
    DISPLAYED_DROPDOWN_DATASET_ENUMS,
    MULTIPLE_LANGUAGE_DATASET_METADATA,
)

logger = logging.getLogger(__name__)


def accept_dataset_metadata_input(
    value: Any, metadata_identifier: str
) -> Tuple[bool, str]:
    logger.debug(f"Received update {value = } for {metadata_identifier = }")
    try:
        if (
            metadata_identifier in MULTIPLE_LANGUAGE_DATASET_METADATA
            and type(value) is str
        ):
            value = find_existing_language_string(
                state.metadata.meta.dataset, value, metadata_identifier
            )

        logger.debug(f"Updating {value = } for {metadata_identifier = }")
        # Update the value in the model
        setattr(
            state.metadata.meta.dataset,
            metadata_identifier,
            value,
        )
    except ValidationError as e:
        show_error = True
        error_explanation = f"`{e}`"
        logger.debug("Caught ValidationError:", exc_info=True)
    else:
        show_error = False
        error_explanation = ""
        logger.debug(f"Successfully updated {metadata_identifier} with {value}")

    return show_error, error_explanation


def update_dataset_metadata_language() -> List[Any]:
    """Return new values for ALL the dataset metadata inputs to allow
    editing of strings in the chosen language"""
    return [
        m.value_getter(state.metadata.meta.dataset, m.identifier)
        for m in DISPLAYED_DATASET_METADATA
    ]


def change_language_dataset_metadata(language):
    update_global_language_state(language)
    return (
        *(
            get_options_for_language(language, e)
            for e in DISPLAYED_DROPDOWN_DATASET_ENUMS
        ),
        update_dataset_metadata_language(),
    )



import re

import dash_bootstrap_components as dbc
from dash import dcc, html


def make_ssb_styled_tab(label: str, content: dbc.Container) -> dbc.Tab:
    return dbc.Tab(
        label=label,
        # Replace all whitespace with dashes
        tab_id=re.sub(r"\s+", "-", label.lower()),
        label_class_name="ssb-tabs navigation-item",
        label_style={"margin-left": "10px", "margin-right": "10px"},
        style={"padding": "4px"},
        children=content,
    )


def make_ssb_warning_alert(
    alert_identifier: str, title: str, content_identifier: str
) -> dbc.Alert:
    return dbc.Alert(
        id=alert_identifier,
        is_open=False,
        dismissable=True,
        fade=True,
        class_name="ssb-dialog warning",
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        width=1,
                        children=[
                            html.Div(
                                className="ssb-dialog warning icon-panel",
                                children=[
                                    html.I(
                                        className="bi bi-exclamation-triangle",
                                    ),
                                ],
                            )
                        ],
                    ),
                    dbc.Col(
                        [
                            html.H5(
                                title,
                            ),
                            dcc.Markdown(
                                id=content_identifier,
                            ),
                        ]
                    ),
                ],
            )
        ],
        color="danger",
    )



import dash_bootstrap_components as dbc
from dash import html
from datadoc.frontend.Builders import make_ssb_warning_alert

dataset_validation_error = make_ssb_warning_alert(
    "dataset-validation-error",
    "Failed validation",
    "dataset-validation-explanation",
)

variables_validation_error = make_ssb_warning_alert(
    "variables-validation-error",
    "Failed validation",
    "variables-validation-explanation",
)

success_toast = dbc.Alert(
    id="success-message",
    is_open=False,
    dismissable=True,
    fade=True,
    class_name="ssb-dialog",
    children=[
        dbc.Row(
            [
                dbc.Col(
                    width=3,
                    children=[
                        html.Div(
                            className="ssb-dialog icon-panel",
                            children=[
                                html.I(
                                    className="bi bi-check-circle",
                                ),
                            ],
                        )
                    ],
                ),
                dbc.Col(
                    [
                        html.H5(
                            "Lagret metadata",
                        ),
                    ]
                ),
            ],
            align="center",
        )
    ],
    color="success",
)



import datadoc.state as state
from dash import Dash
from datadoc.app import build_app
from datadoc.backend.DataDocMetadata import DataDocMetadata

from .utils import TEST_PARQUET_FILEPATH


def test_build_app():
    state.metadata = DataDocMetadata(TEST_PARQUET_FILEPATH)
    app = build_app(Dash)
    assert app.config["name"] == "Datadoc"
    assert len(app.callback_map.items()) > 0



from datadoc.frontend.fields.DisplayDataset import DISPLAY_DATASET, DatasetIdentifiers
from datadoc.frontend.fields.DisplayVariables import (
    DISPLAY_VARIABLES,
    VariableIdentifiers,
)
from datadoc_model.Model import DataDocDataSet, DataDocVariable


def test_dataset_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in DatasetIdentifiers] == list(DataDocDataSet().dict().keys())
    assert list(DatasetIdentifiers) == list(DISPLAY_DATASET.keys())


def test_variables_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in VariableIdentifiers] == list(
        DataDocVariable().dict().keys()
    )
    assert list(VariableIdentifiers) == list(DISPLAY_VARIABLES.keys())



import logging
from typing import Dict, List, Optional, Tuple

from datadoc_model.Enums import SupportedLanguages
from pydantic import ValidationError

import datadoc.state as state
from datadoc.frontend.callbacks.utils import (
    find_existing_language_string,
    get_options_for_language,
)
from datadoc.frontend.fields.DisplayVariables import (
    DISPLAYED_DROPDOWN_VARIABLES_METADATA,
    DISPLAYED_DROPDOWN_VARIABLES_TYPES,
    MULTIPLE_LANGUAGE_VARIABLES_METADATA,
    VariableIdentifiers,
)
from datadoc.utils import get_display_values

logger = logging.getLogger(__name__)


def get_boolean_options_for_language(language: SupportedLanguages):
    true_labels = {
        SupportedLanguages.ENGLISH: "Yes",
        SupportedLanguages.NORSK_NYNORSK: "Ja",
        SupportedLanguages.NORSK_BOKMÅL: "Ja",
    }
    false_labels = {
        SupportedLanguages.ENGLISH: "No",
        SupportedLanguages.NORSK_NYNORSK: "Nei",
        SupportedLanguages.NORSK_BOKMÅL: "Nei",
    }
    return [
        {
            "label": f"{true_labels[language]}",
            "value": True,
        },
        {
            "label": f"{false_labels[language]}",
            "value": False,
        },
    ]


def accept_variable_metadata_input(
    data: List[Dict], data_previous: List[Dict]
) -> Tuple[List[Dict], bool, str]:
    updated_row_id: str = ""
    updated_column_id: str = ""
    new_value: Optional[str] = None
    show_error = False
    error_explanation = ""
    output_data = data
    update_diff = []
    # What has changed?
    for i in range(len(data)):
        update_diff = list(data[i].items() - data_previous[i].items())
        if update_diff:
            updated_row_id = data[i][VariableIdentifiers.SHORT_NAME.value]
            updated_column_id = update_diff[-1][0]
            new_value = update_diff[-1][-1]
            break  # We're only interested in one change so we break here

    if update_diff:
        try:
            if updated_column_id in MULTIPLE_LANGUAGE_VARIABLES_METADATA:
                if type(new_value) is str:
                    new_value = find_existing_language_string(
                        state.metadata.variables_lookup[updated_row_id],
                        new_value,
                        updated_column_id,
                    )
                elif new_value is None:
                    # This edge case occurs when the user removes the text in an input field
                    # We want to ensure we only remove the content for the current language,
                    # not create a new blank object!
                    new_value = find_existing_language_string(
                        state.metadata.variables_lookup[updated_row_id],
                        "",
                        updated_column_id,
                    )

            logger.debug(
                f"Row: {updated_row_id} Column: {updated_column_id} New value: {new_value}"
            )
            # Write the value to the variables structure
            setattr(
                state.metadata.variables_lookup[updated_row_id],
                updated_column_id,
                new_value,
            )
        except ValidationError as e:
            show_error = True
            error_explanation = f"`{e}`"
            output_data = data_previous
            logger.debug("Caught ValidationError:", exc_info=True)
        else:
            logger.debug(f"Successfully updated {updated_row_id} with {new_value}")

    return output_data, show_error, error_explanation


def update_variable_table_dropdown_options_for_language(
    language: SupportedLanguages,
) -> Dict[str, Dict[str, List[Dict[str, str]]]]:
    """Retrieves enum options for dropdowns in the Datatable. Handles the
    special case of boolean values which we represent in the Datatable
    with a Dropdown but they're not backed by an Enum.

    Example return structure:
        {'data_type': {'options': [{'label': 'TEKST', 'value': 'STRING'},
                                {'label': 'HELTALL', 'value': 'INTEGER'},
                                {'label': 'DESIMALTALL', 'value': 'FLOAT'},
                                {'label': 'DATOTID', 'value': 'DATETIME'},
                                {'label': 'BOOLSK', 'value': 'BOOLEAN'}]},
        'direct_person_identifying': {'options': [{'label': 'Ja', 'value': True},
                                                {'label': 'Nei', 'value': False}]},
        'temporality_type': {'options': [{'label': 'FAST', 'value': 'FIXED'},
            ...
        }
    """
    options = []
    for field_type in DISPLAYED_DROPDOWN_VARIABLES_TYPES:
        value = (
            get_boolean_options_for_language(language)
            if field_type is bool
            else get_options_for_language(language, field_type)
        )
        options.append({"options": value})
    return dict(zip(DISPLAYED_DROPDOWN_VARIABLES_METADATA, options))


def update_variable_table_language(
    data: List[Dict],
    language: SupportedLanguages,
) -> Tuple[List[Dict], bool, str]:
    state.current_metadata_language = language
    new_data = []
    for row in data:
        new_data.append(
            get_display_values(
                state.metadata.variables_lookup[
                    row[VariableIdentifiers.SHORT_NAME.value]
                ],
                state.current_metadata_language,
            )
        )
    logger.debug(f"Updated variable table language: {language.name}")
    return new_data, False, ""



import pathlib
from abc import ABC, abstractmethod
from typing import List, Optional, TypeVar

import pandas as pd
import pyarrow.parquet as pq
from datadoc_model.Enums import Datatype
from datadoc_model.LanguageStrings import LanguageStrings
from datadoc_model.Model import DataDocVariable

from datadoc import state
from datadoc.backend.StorageAdapter import StorageAdapter

TDatasetParser = TypeVar("TDatasetParser", bound="DatasetParser")

KNOWN_INTEGER_TYPES = (
    "int",
    "int_",
    "int8",
    "int16",
    "int32",
    "int64",
    "integer",
    "long",
    "uint",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
)

KNOWN_FLOAT_TYPES = (
    "double",
    "float",
    "float_",
    "float16",
    "float32",
    "float64",
    "decimal",
    "number",
    "numeric",
    "num",
)

KNOWN_STRING_TYPES = (
    "string",
    "str",
    "char",
    "varchar",
    "varchar2",
    "text",
    "txt",
    "bytes",
)

KNOWN_DATETIME_TYPES = (
    "timestamp",
    "timestamp[us]",
    "timestamp[ns]",
    "datetime64",
    " datetime64[ns]",
    " datetime64[us]",
    "date",
    "datetime",
    "time",
)

KNOWN_BOOLEAN_TYPES = ("bool", "bool_", "boolean")


class DatasetParser(ABC):
    def __init__(self, dataset: str):
        self.dataset: StorageAdapter = StorageAdapter.for_path(dataset)

    @staticmethod
    def for_file(dataset: str) -> TDatasetParser:
        """Factory method to return the correct subclass based on the given dataset file"""
        supported_file_types = {
            "parquet": DatasetParserParquet,
            "sas7bdat": DatasetParserSas7Bdat,
        }
        file_type = "Unknown"
        try:
            file_type = str(pathlib.Path(dataset)).lower().split(".")[-1]
            # Extract the appropriate reader class from the SUPPORTED_FILE_TYPES dict and return an instance of it
            reader = supported_file_types[file_type](dataset)
        except IndexError as e:
            # Thrown when just one element is returned from split, meaning there is no file extension supplied
            raise FileNotFoundError(
                f"Could not recognise file type for provided {dataset = }. Supported file types are: {', '.join(supported_file_types.keys())}"
            ) from e
        except KeyError as e:
            # In this case the file type is not supported, so we throw a helpful exception
            raise NotImplementedError(
                f"{file_type = } is not supported. Please open one of the following supported files types: {', '.join(supported_file_types.keys())} or contact the maintainers to request support."
            ) from e
        else:
            return reader

    @staticmethod
    def transform_data_type(data_type: str) -> Optional[Datatype]:
        v_data_type = data_type.lower()
        if v_data_type in KNOWN_INTEGER_TYPES:
            return Datatype.INTEGER
        elif v_data_type in KNOWN_FLOAT_TYPES:
            return Datatype.FLOAT
        elif v_data_type in KNOWN_STRING_TYPES:
            return Datatype.STRING
        elif v_data_type in KNOWN_DATETIME_TYPES:
            return Datatype.DATETIME
        elif v_data_type in KNOWN_BOOLEAN_TYPES:
            return Datatype.BOOLEAN
        else:
            # Unknown data type. There's no need to throw an exception here,
            # the user can still define the data type manually in the GUI
            return None

    @abstractmethod
    def get_fields(self) -> List[DataDocVariable]:
        """Abstract method, must be implemented by subclasses"""


class DatasetParserParquet(DatasetParser):
    def __init__(self, dataset: str):
        super().__init__(dataset)

    def get_fields(self) -> List[DataDocVariable]:
        fields = []
        with self.dataset.open(mode="rb") as f:
            data_table = pq.read_table(f)
            for data_field in data_table.schema:
                fields.append(
                    DataDocVariable(
                        short_name=data_field.name,
                        data_type=self.transform_data_type(str(data_field.type)),
                    )
                )
        return fields


class DatasetParserSas7Bdat(DatasetParser):
    def __init__(self, dataset: str):
        super().__init__(dataset)

    def get_fields(self) -> List[DataDocVariable]:
        fields = []
        with self.dataset.open(mode="rb") as f:
            # Use an iterator to avoid reading in the entire dataset
            sas_reader = pd.read_sas(f, format="sas7bdat", iterator=True)

            # Get the first row from the iterator
            row = next(sas_reader)

        # Get all the values from the row and loop through them
        for i, v in enumerate(row.values.tolist()[0]):
            fields.append(
                DataDocVariable(
                    short_name=sas_reader.columns[i].name,
                    # Assume labels are defined in the default language (NORSK_BOKMÅL)
                    # If this is not correct, the user may fix it via the UI
                    name=LanguageStrings(
                        **{state.current_metadata_language: sas_reader.columns[i].label}
                    ),
                    # Access the python type for the value and transform it to a DataDoc Data type
                    data_type=self.transform_data_type(type(v).__name__.lower()),
                )
            )

        return fields



import pytest

from datadoc.tests.utils import TEST_BUCKET_PARQUET_FILEPATH, TEST_PARQUET_FILEPATH

from ..backend.StorageAdapter import StorageAdapter


@pytest.fixture()
def mock_gcsfs_open(mocker):
    mock = mocker.patch("gcsfs.GCSFileSystem.open")
    return mock


@pytest.fixture()
def mock_gcsfs_exists(mocker):
    mock = mocker.patch("gcsfs.GCSFileSystem.exists")
    mock.return_value = True
    return mock


@pytest.fixture()
def mock_pathlib_write_text(mocker):
    mock = mocker.patch("pathlib.Path.write_text")
    return mock


@pytest.fixture
def local_parquet_file(mock_pathlib_write_text):
    return StorageAdapter.for_path(TEST_PARQUET_FILEPATH)


@pytest.fixture
def bucket_object_parquet_file(mock_gcsfs_open, mock_gcsfs_exists):
    return StorageAdapter.for_path(TEST_BUCKET_PARQUET_FILEPATH)



import argparse
import logging
import os
from typing import Type

import dash_bootstrap_components as dbc
from dash import Dash

import datadoc.state as state
from datadoc.backend.DataDocMetadata import DataDocMetadata
from datadoc.frontend.callbacks.register import register_callbacks
from datadoc.frontend.components.Alerts import (
    dataset_validation_error,
    success_toast,
    variables_validation_error,
)
from datadoc.frontend.components.DatasetTab import get_dataset_tab
from datadoc.frontend.components.HeaderBars import (
    get_controls_bar,
    header,
    progress_bar,
)
from datadoc.frontend.components.VariablesTab import get_variables_tab
from datadoc.utils import running_in_notebook

logger = logging.getLogger(__name__)

NAME = "Datadoc"


def build_app(dash_class: Type[Dash]) -> Dash:

    app = dash_class(
        name=NAME,
        title=NAME,
        external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP],
        assets_folder=f"{os.path.dirname(__file__)}/assets",
    )

    app.layout = dbc.Container(
        style={"padding": "4px"},
        children=[
            header,
            progress_bar,
            get_controls_bar(),
            dbc.CardBody(
                style={"padding": "4px"},
                children=[
                    dbc.Tabs(
                        id="tabs",
                        class_name="ssb-tabs",
                        children=[
                            get_dataset_tab(),
                            get_variables_tab(),
                        ],
                    ),
                ],
            ),
            variables_validation_error,
            dataset_validation_error,
            success_toast,
        ],
    )

    register_callbacks(app)

    return app


def main(dataset_path: str = None):
    logging.basicConfig(level=logging.DEBUG)
    if dataset_path is None:
        # Get the supplied command line argument
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "dataset_path",
            help="Specify the path to a dataset",
            nargs="?",
            default=f"{os.path.dirname(__file__)}/../klargjorte_data/person_data_v1.parquet",
        )
        dataset = parser.parse_args().dataset_path
    else:
        dataset = dataset_path

    state.metadata = DataDocMetadata(dataset)

    if running_in_notebook():
        logging.basicConfig(level=logging.INFO)
        from jupyter_dash import JupyterDash

        JupyterDash.infer_jupyter_proxy_config()
        app = build_app(JupyterDash)
        app.run_server(mode="jupyterlab")
    else:
        # Assume running in server mode is better (largely for development purposes)
        logger.debug("Starting in development mode")
        app = build_app(Dash)
        app.run(debug=True, use_reloader=False)


if __name__ == "__main__":
    main()



from typing import Any, Dict, List, Tuple

from dash import ALL, Dash, Input, Output, ctx
from datadoc_model.Enums import SupportedLanguages

import datadoc.state as state
from datadoc.frontend.callbacks.dataset import (
    accept_dataset_metadata_input,
    change_language_dataset_metadata,
)
from datadoc.frontend.callbacks.variables import (
    accept_variable_metadata_input,
    update_variable_table_dropdown_options_for_language,
    update_variable_table_language,
)
from datadoc.frontend.components.DatasetTab import DATASET_METADATA_INPUT
from datadoc.frontend.fields.DisplayDataset import DISPLAYED_DROPDOWN_DATASET_METADATA

# Avoid implementing callbacks here.
#
# The Dash Inputs and Outputs etc. shall be defined here, but the
# implementations should be in other functions, to enable unit testing


def register_callbacks(app: Dash) -> None:
    @app.callback(
        Output("progress-bar", "value"),
        Output("progress-bar", "label"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        Input("variables-table", "data"),
    )
    def callback_update_progress(value, data) -> Tuple[int, str]:
        completion = state.metadata.percent_complete
        return completion, f"{completion}%"

    @app.callback(
        Output("success-message", "is_open"),
        Input("save-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(n_clicks):
        if n_clicks and n_clicks > 0:
            # Write the final completion percentage to the model
            state.metadata.meta.percentage_complete = state.metadata.percent_complete
            state.metadata.write_metadata_document()
            return True
        else:
            return False

    @app.callback(
        *[
            Output(
                {
                    "type": DATASET_METADATA_INPUT,
                    "id": m.identifier,
                },
                "options",
            )
            for m in DISPLAYED_DROPDOWN_DATASET_METADATA
        ],
        Output(
            {"type": DATASET_METADATA_INPUT, "id": ALL},
            "value",
        ),
        Input("language-dropdown", "value"),
    )
    def callback_change_language_dataset_metadata(language: str):
        return change_language_dataset_metadata(SupportedLanguages(language))

    @app.callback(
        Output("dataset-validation-error", "is_open"),
        Output("dataset-validation-explanation", "children"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(value: Any) -> Tuple[bool, str]:
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"], ctx.triggered_id["id"]
        )

    @app.callback(
        Output("variables-table", "data"),
        Output("variables-validation-error", "is_open"),
        Output("variables-validation-explanation", "children"),
        Input("variables-table", "data"),
        Input("variables-table", "data_previous"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_variable_table(
        data: List[Dict], data_previous: List[Dict], language: str
    ) -> Tuple[List[Dict], bool, str]:
        if ctx.triggered_id == "language-dropdown":
            return update_variable_table_language(data, SupportedLanguages(language))
        else:
            return accept_variable_metadata_input(data, data_previous)

    @app.callback(
        Output("variables-table", "dropdown"),
        Input("language-dropdown", "value"),
    )
    def callback_variable_table_dropdown_options(language: str):
        language = SupportedLanguages(language)
        return update_variable_table_dropdown_options_for_language(language)



import os
import shutil
from copy import copy
from pathlib import PurePath
from typing import List, Tuple

import pytest
from datadoc_model import Enums
from datadoc_model.Enums import DatasetState
from datadoc_model.Model import DataDocDataSet, DataDocVariable, MetadataDocument

from datadoc.backend.DataDocMetadata import DataDocMetadata

from .utils import (
    TEST_EXISTING_METADATA_FILE_NAME,
    TEST_EXISTING_METADATA_FILEPATH,
    TEST_PARQUET_FILEPATH,
    TEST_RESOURCES_DIRECTORY,
)


@pytest.fixture
def metadata():
    yield DataDocMetadata(TEST_PARQUET_FILEPATH)


def make_paths() -> List[Tuple[str, DatasetState]]:
    split_path = list(PurePath(TEST_PARQUET_FILEPATH).parts)
    initial_data = [
        ("kildedata", DatasetState.SOURCE_DATA),
        ("inndata", DatasetState.INPUT_DATA),
        ("roskildedata/klargjorte-data", DatasetState.PROCESSED_DATA),
        ("klargjorte_data", DatasetState.PROCESSED_DATA),
        ("klargjorte-data", DatasetState.PROCESSED_DATA),
        ("statistikk", DatasetState.STATISTIC),
        ("", None),
    ]
    test_data = []

    # Construct paths with each of the potential options in them
    for to_insert, state in initial_data:
        new_path = copy(split_path)
        new_path.insert(-2, to_insert)
        new_path = PurePath("").joinpath(*new_path)
        test_data.append((str(new_path), state))

    return test_data


@pytest.fixture
def remove_document_file() -> None:
    yield None  # Dummy value, No need to return anything in particular here
    os.remove(os.path.join(TEST_RESOURCES_DIRECTORY, TEST_EXISTING_METADATA_FILE_NAME))


@pytest.fixture
def existing_metadata_file():
    # Setup by copying the file into the relevant directory
    shutil.copy(TEST_EXISTING_METADATA_FILEPATH, TEST_RESOURCES_DIRECTORY)
    yield None  # Dummy value, No need to return anything in particular here


@pytest.mark.parametrize(("path", "expected_result"), make_paths())
def test_get_dataset_state(
    path: str, expected_result: DatasetState, metadata: DataDocMetadata
):
    actual_state = metadata.get_dataset_state(path)
    assert actual_state == expected_result


def test_get_dataset_state_none(metadata):
    assert metadata.get_dataset_state(None) is None


def test_existing_metadata_file(existing_metadata_file, metadata, remove_document_file):
    assert metadata.meta.dataset.name.en == "successfully_read_existing_file"


def test_metadata_document_percent_complete(metadata):
    dataset = DataDocDataSet(dataset_state=Enums.DatasetState.OUTPUT_DATA)
    variable_1 = DataDocVariable(data_type=Enums.Datatype.BOOLEAN)
    variable_2 = DataDocVariable(data_type=Enums.Datatype.INTEGER)
    document = MetadataDocument(
        percentage_complete=0,
        document_version=1,
        dataset=dataset,
        variables=[variable_1, variable_2],
    )
    metadata.meta = document

    assert metadata.percent_complete == 11


def test_get_dataset_version(metadata: DataDocMetadata):
    assert metadata.get_dataset_version(metadata.short_name) == "1"


def test_get_dataset_version_unknown(metadata: DataDocMetadata):
    assert metadata.get_dataset_version("person_data.parquet") is None


def test_write_metadata_document(metadata, remove_document_file):
    metadata.write_metadata_document()
    assert os.path.exists(
        os.path.join(TEST_RESOURCES_DIRECTORY, TEST_EXISTING_METADATA_FILE_NAME)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datadoc',
 'datadoc.backend',
 'datadoc.frontend',
 'datadoc.frontend.callbacks',
 'datadoc.frontend.components',
 'datadoc.frontend.fields',
 'datadoc.tests']

package_data = \
{'': ['*'],
 'datadoc': ['assets/*'],
 'datadoc.tests': ['resources/*', 'resources/existing_metadata_file/*']}

install_requires = \
['dash-bootstrap-components>=1.1.0,<2.0.0',
 'dash>=2.4.1,<3.0.0',
 'jupyter-dash>=0.4.2,<0.5.0',
 'pandas>=1.4.2,<2.0.0',
 'pyarrow>=8.0.0,<9.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'ssb-datadoc-model==0.1.3']

extras_require = \
{'gcs': ['dapla-toolbelt>=1.3.1,<2.0.0', 'gcsfs>=2022.7.1,<2023.0.0']}

entry_points = \
{'console_scripts': ['datadoc = datadoc.app:main']}

setup_kwargs = {
    'name': 'ssb-datadoc',
    'version': '0.1.5',
    'description': "Document dataset metadata. For use in Statistics Norway's metadata system.",
    'long_description': '# Datadoc\n\n![Datadoc Unit tests](https://github.com/statisticsnorway/datadoc/actions/workflows/unit-tests.yml/badge.svg) ![Code coverage](https://img.shields.io/endpoint?url=https://gist.githubusercontent.com/mmwinther/0c0c5bdfc360b59254f2c32d65914025/raw/pytest-coverage-badge-datadoc.json) ![PyPI version](https://img.shields.io/pypi/v/ssb-datadoc) ![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)\n\nDocument datasets in Statistics Norway\n\n## Usage\n\n![DataDoc in use](./doc/change-language-example.gif)\n\n### From Jupyter\n\n1. Open <https://jupyter.dapla-staging.ssb.no> or another Jupyter Lab environment\n1. Run `pip install ssb-datadoc[gcs]` in the terminal\n1. Upload a dataset to your Jupyter server (e.g. <https://github.com/statisticsnorway/datadoc/blob/master/klargjorte_data/person_data_v1.parquet>)\n1. Run `from datadoc import main; main("./person_data_v1.parquet")` in a notebook\n1. Datadoc will open in a new tab\n\n## Contributing\n\n### Prerequisites\n\n- Python >3.8 (3.10 is preferred)\n- Poetry, install via `curl -sSL https://install.python-poetry.org | python3 -`\n\n### Dependency Management\n\nPoetry is used for dependency management. [Poe the Poet](https://github.com/nat-n/poethepoet) is used for running poe tasks within poetry\'s virtualenv. Upon cloning this project first install necessary dependencies, then run the tests to verify everything is working.\n\n#### Install all dependencies\n\n```shell\npoetry install --all-extras\n```\n\n### Add dependencies\n\n#### Main\n\n```shell\npoetry add <python package name>\n```\n\n#### Dev\n\n```shell\npoetry add --group dev <python package name>\n```\n\n### Run tests\n\n```shell\npoetry run poe test\n```\n\n### Run project locally\n\nTo run the project locally:\n\n```shell\npoetry run poe datadoc "gs://ssb-staging-dapla-felles-data-delt/datadoc/klargjorte_data/person_data_v1.parquet"\n```\n\n### Run project locally in Jupyter\n\nTo run the project locally in Jupyter run:\n\n```shell\npoetry run poe jupyter\n```\n\nA Jupyter instance should open in your browser. Open and run the cells in the `.ipynb` file to demo datadoc.\n\n### Bump version\n\n```shell\npoetry run poe bump-patch-version\n```\n\n> :warning: Run this on the default branch\n\nThis command will:\n\n1. Increment version strings in files\n1. Commit the changes\n1. Tag the commit with the new version\n\nThen just run `git push origin --tags` to push the changes and trigger the release process.\n',
    'author': 'Statistics Norway',
    'author_email': 'mmw@ssb.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/statisticsnorway/datadoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import pytest
from datadoc_model.Enums import SupportedLanguages

from datadoc import state


@pytest.fixture()
def clear_state():
    state.metadata = None
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL



import dash_bootstrap_components as dbc
import datadoc.state as state
from dash import dcc, html
from datadoc_model.Enums import SupportedLanguages

COLORS = {"dark_1": "#F0F8F9", "green_1": "#ECFEED", "green_4": "#00824D"}

header = dbc.CardBody(
    dbc.Row(
        children=[
            html.H1("DataDoc", className="ssb-title", style={"color": "white"}),
        ],
    ),
    style={"backgroundColor": COLORS["green_4"]},
)

progress_bar = dbc.CardBody(
    style={"padding": "4px"},
    children=[dbc.Progress(id="progress-bar", color=COLORS["green_4"], value=40)],
)


def get_controls_bar() -> dbc.CardBody:
    return dbc.CardBody(
        style={"padding": "4px"},
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Button(
                            [
                                html.I(
                                    className="bi bi-save",
                                    style={"padding-right": "10px"},
                                ),
                                "   Lagre",
                            ],
                            class_name="ssb-btn primary-btn",
                            id="save-button",
                        ),
                    ),
                    dbc.Col(
                        dcc.Dropdown(
                            id="language-dropdown",
                            searchable=False,
                            value=state.current_metadata_language.value,
                            className="ssb-dropdown",
                            options=[
                                {"label": i.name, "value": i.value}
                                for i in SupportedLanguages
                            ],
                        ),
                        align="end",
                        width="auto",
                    ),
                ]
            )
        ],
    )



import logging
import os
import pathlib
from io import IOBase, TextIOWrapper
from typing import Protocol
from urllib.parse import urlsplit, urlunsplit

GCS_PROTOCOL_PREFIX = "gs://"

logger = logging.getLogger(__name__)


class GCSObject:
    def __init__(self, path: str):
        self._url = urlsplit(path)
        try:
            from dapla import AuthClient, FileClient
            if AuthClient.is_ready():
                # Running on Dapla, rely on dapla-toolbelt for auth
                self.fs = FileClient.get_gcs_file_system()
            else:
                # All other environments, rely on Standard Google credential system
                # If this doesn't work for you, try running the following commands:
                #
                # gcloud auth application-default revoke
                # gcloud auth application-default login
                from gcsfs import GCSFileSystem
                self.fs = GCSFileSystem()

        except ImportError:
            msg = f"Missing support for GCS. Install datadoc with 'pip install ssb-datadoc[gcs]'"
            raise ImportError(msg)

    def _rebuild_url(self, new_path: str) -> str:
        return urlunsplit((self._url.scheme, self._url.netloc, new_path, None, None))

    def open(self, **kwargs) -> IOBase:
        return self.fs.open(self.location, **kwargs)

    def parent(self) -> str:
        parent = os.path.dirname(self._url.path)
        return self._rebuild_url(parent)

    def joinpath(self, part):
        """Modify the path in place"""
        self._url = urlsplit(self._rebuild_url(os.path.join(self._url.path, part)))

    def exists(self) -> bool:
        return self.fs.exists(self.location)

    def write_text(self, text: str) -> None:
        f: TextIOWrapper
        with self.fs.open(self.location, mode="w") as f:
            f.write(text)

    @property
    def location(self) -> str:
        return urlunsplit(self._url)


class LocalFile:
    def __init__(self, path):
        self._path_object: pathlib.Path = pathlib.Path(path)

    def open(self, **kwargs) -> IOBase:
        return open(str(self._path_object), **kwargs)

    def parent(self) -> str:
        return str(self._path_object.resolve().parent)

    def joinpath(self, part):
        """Modify the path in place"""
        self._path_object = self._path_object.joinpath(part)

    def exists(self) -> bool:
        return self._path_object.exists()

    def write_text(self, text: str) -> None:
        self._path_object.write_text(text, encoding="utf-8")

    @property
    def location(self) -> str:
        return str(self._path_object)


class StorageAdapter(Protocol):
    @staticmethod
    def for_path(path: str):
        """
        Return a concrete class implementing this Protocol based on the structure of the path.
        """
        if path.startswith(GCS_PROTOCOL_PREFIX):
            return GCSObject(path)
        else:
            return LocalFile(path)

    def open(self, **kwargs) -> IOBase:
        ...

    def parent(self) -> str:
        ...

    def joinpath(self, part: str) -> None:
        ...

    def exists(self) -> bool:
        ...

    def write_text(self, text: str) -> None:
        ...

    @property
    def location(self) -> str:
        ...



import pathlib

import pytest

from datadoc.tests.storage_adapter_fixtures import *  # noqa pytest magic
from datadoc.tests.utils import TEST_BUCKET_PARQUET_FILEPATH, TEST_PARQUET_FILEPATH

from ..backend.StorageAdapter import GCSObject, LocalFile, StorageAdapter


@pytest.mark.parametrize(
    ("file", "expected_class"),
    [("local_parquet_file", LocalFile), ("bucket_object_parquet_file", GCSObject)],
)
def test_factory(file: StorageAdapter, expected_class, request):
    # Ugly pytest magic to get the actual fixture out
    file = request.getfixturevalue(file)
    assert isinstance(file, expected_class)


@pytest.mark.parametrize("file", ["local_parquet_file", "bucket_object_parquet_file"])
def test_open(file: StorageAdapter, request):
    # Ugly pytest magic to get the actual fixture out
    file = request.getfixturevalue(file)
    with file.open() as file_handle:
        assert file_handle.readable()


@pytest.mark.parametrize(
    ("file", "expected_parent"),
    [
        (
            "local_parquet_file",
            str(pathlib.Path(TEST_PARQUET_FILEPATH).parent.absolute()),
        ),
        (
            "bucket_object_parquet_file",
            "/".join(TEST_BUCKET_PARQUET_FILEPATH.split("/")[:-1]),
        ),
    ],
)
def test_parent(file: StorageAdapter, expected_parent: str, request):
    # Ugly pytest magic to get the actual fixture out
    file = request.getfixturevalue(file)
    assert file.parent() == expected_parent


@pytest.mark.parametrize(
    ("file", "expected"),
    [
        (
            "local_parquet_file",
            "/".join([TEST_PARQUET_FILEPATH, "extra"]),
        ),
        (
            "bucket_object_parquet_file",
            "/".join([TEST_BUCKET_PARQUET_FILEPATH, "extra"]),
        ),
    ],
)
def test_joinpath(file: StorageAdapter, expected: str, request):
    # Ugly pytest magic to get the actual fixture out
    file = request.getfixturevalue(file)
    file.joinpath("extra")
    assert file.location == expected


@pytest.mark.parametrize("file", ["local_parquet_file", "bucket_object_parquet_file"])
def test_exists(file: StorageAdapter, request):
    # Ugly pytest magic to get the actual fixture out
    file = request.getfixturevalue(file)
    assert file.exists()


def test_write_text_local_file(local_parquet_file: StorageAdapter, request):
    local_parquet_file.write_text("12345")
    mock = request.getfixturevalue("mock_pathlib_write_text")
    mock.assert_called_once_with("12345", encoding="utf-8")


# Currently no test for writing to GCS
# Attempts to mock are failing with TypeError: cannot set 'write' attribute of immutable type '_io.BufferedWriter'
